/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "buffers.c" - Bufferized input/output: here, and only here,    *
 *                       the FAT driver interfaces with the disk driver   *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"
#include "ide.h"


/* Searches for a volume's buffer contaning the specified sector. */
/* Returns the buffer number on success, or -1 on failure.        */
/* Called by read_from_buffer.                                    */
static int search_for_buffer(tFatVolume *V, DWORD Sector, unsigned int *SectorOffset)
{
  int k;
  for (k = 0; k < FAT_MAX_BUFFERS; k++)
    if( (Sector >= V->Buffer[k].StartingSector) &&
        (Sector < (V->Buffer[k].StartingSector+V->SectorsInBuffer))){
      *SectorOffset = (unsigned int) ((Sector-V->Buffer[k].StartingSector) << 9);
      return k;
    }
  return -1;
}


/* Returns the number of the least recently used volume's buffer. */
/* Called by read_from_buffer.                                    */
static int find_least_recently_used(tFatVolume *V)
{
  int   k, Lru = 0;
  DWORD MinAccess = V->BufferAccess;
  for (k = 0; k < FAT_MAX_BUFFERS; k++)
    if (V->Buffer[k].LastAccess < MinAccess)
    {
      MinAccess = V->Buffer[k].LastAccess;
      Lru       = k;
    }
  return Lru;
}


/* Writes the content of the specified buffer in the hosting block */
/* device, marking the buffer as not dirty.                        */
/* Returns 0 on success, or a negative error code on failure.      */
/* Called by flush_all_buffers and read_from_buffer.               */
static int flush_buffer(tFatVolume *V, int NumBuf)
{
  int Res;
  if (V->Buffer[NumBuf].Flags & FAT_BUFFER_DIRTY)
  {
    unsigned long lba;
    unsigned int  c, h, s;
if (V->Buffer[NumBuf].StartingSector == 0) {
  printf("Writing Sector 0 not allowed!!");
  return 0;
}
    lba = V->Buffer[NumBuf].StartingSector+V->PartStart;
    LBA_to_PCHS(lba, &c, &h, &s);
    Res = IDE_Write(c, h, s, V->SectorsInBuffer, V->Buffer[NumBuf].Data);
//    Res = V->Blk->write(V->Blk->P, V->Buffer[NumBuf].StartingSector, 1,
//                        V->Buffer[NumBuf].Data);
    if (Res == 0) return FAT_RET_SECTOR_ERROR;
    V->Buffer[NumBuf].Flags &= ~FAT_BUFFER_DIRTY;
  }
  return 0;
}


/* Flushes all buffers of the specified volume, calling flush_buffer. */
/* Returns 0 on success, or a negative error code on failure.         */
/* Called by fat_fflush in open.c.                                    */
int flush_all_buffers(tFatVolume *V)
{
  int k, Res;
  for (k = 0; k < FAT_MAX_BUFFERS; k++)
  {
    Res = flush_buffer(V, k);
    if (Res) return Res;
  }
  return 0;
}


/* Performs a bufferized read from the hosting block device.      */
/* If the required sector is not already bufferized, flushes the  */
/* least recently used buffers and uses it to load in the sector. */
/* Puts the buffer number in NumBuf.                              */
/* Returns zero on success, or a negative error code on failure.  */
/* Called by fat??_read_entry and fat??_write_entry in fat??.c,   */
/* fat_read in read.c and fat_write in write.c.                   */
int read_from_buffer(tFatVolume *V, DWORD Sector, unsigned int *SectorOffset)
{
  int Res;

 #ifdef FATBUFFERS

  int NumBuf = search_for_buffer(V, Sector, SectorOffset);
  if (NumBuf == -1)
  {
    V->BufferMiss++;
    NumBuf = find_least_recently_used(V);
    Res = flush_buffer(V, NumBuf);
    if (Res < 0) return Res;

    unsigned long lba;
    unsigned int  c, h, s;
    lba = Sector+V->PartStart;
    LBA_to_PCHS(lba, &c, &h, &s);
    Res = IDE_Read(c, h, s, V->SectorsInBuffer, V->Buffer[NumBuf].Data);
//    Res = V->Blk->read(V->Blk->P, Sector, 1, V->Buffer[NumBuf].Data);
    if (Res == 0) return FAT_RET_SECTOR_ERROR;
    V->Buffer[NumBuf].StartingSector = Sector;
    *SectorOffset = 0;
  }
  else V->BufferHit++;
  V->Buffer[NumBuf].LastAccess = ++V->BufferAccess;
  return NumBuf;

 #else

  Res = V->Blk->read(V->Blk->P, Sector, 1, V->Buffer[0].Data);
  if (Res) return FAT_ERROR_SECTOR;
  return 0;

 #endif
}
