/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "direntry.c" - Internal services to allocate or free directory *
 *                        entries for long or short file names            *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"



/* Searches an open directory for NumEntries free consecutive entries. */
/* On success, returns 0 and fills DirOffset with the offset of the    */
/* first free entry. Returns a negative error code on failure.         */
/* Note that error FAT_ERROR_DIRECTORY_FULL means that no free entries */
/* are available for that directory extension, thus the directory must */
/* be extended (if possible) to find room.                             */
/* Called by allocate_sfn_dir_entry and allocate_lfn_dir_entries.      */
int find_empty_dir_entries(tFatVolume *V, int Fd, int NumEntries,
                           DWORD *EntryOffset)
{
  int          Found = 0; /* Number of consecutive free entries found */
  int          Res;
  DWORD        NumRead;
  tFatDirEntry Buffer;

  if (!(V->Files[Fd].DirEntry.Attr & FD32_ATTR_DIRECTORY))
    return FAT_RET_NOT_A_DIRECTORY;

  /* Scans the whole directory for free entries */
  Res = fat_read(V, Fd, (BYTE *) &Buffer, 32, &NumRead);
  if (Res) return Res;
  while (NumRead != 0)
  {
    if ((Buffer.Name[0] == FAT_DIR_FREE)
     || (Buffer.Name[0] == FAT_DIR_ENDOFDIR))
    {
      if (Found > 0) Found++;
      else
      {
        Found      = 1;
        *EntryOffset = V->Files[Fd].TargetPos - NumRead;
      }
    }
    else Found = 0;
    if (Found == NumEntries) return 0;
    Res = fat_read(V, Fd, (BYTE *) &Buffer, 32, &NumRead);
    if (Res) return Res;
  }
  return FAT_RET_DIRECTORY_FULL;
}


/* Marks as free the specified entries of an open directory. */
/* DirEntryOffset is the byte offset of the short name entry */
/* to be freed, preceded by DirLfnEntries LFN slots.         */
/* Called by fat_unlink (unlink.c) and fat_rename (rename.c) */
int free_dir_entries(tFatVolume *V, int Fd, DWORD EntryOffset, int LfnEntries)
{
  int  Res, k;
  BYTE FirstChar = FAT_DIR_FREE;
  
  /* There are DirLfnEntries+1 entries starting at offset */
  /* DirEntryOffset - DirLfnEntries * 32.                 */
  for (k = LfnEntries; k >= 0; k--)
  {
    V->Files[Fd].TargetPos = EntryOffset - k * 32;
    Res = fat_write(V, Fd, &FirstChar, 1, NULL);
    if (Res) return Res;
  }
  return 0;
}


/* Allocates one entry of the specified open directory and fills it    */
/* with the directory entry D.                                         */
/* On success, returns 0 and fills EntryOffset with the byte offset    */
/* of the allocated entry. Returns a negative error code on failure.   */
/* Called by allocate_lfn_dir_entry (if not LFN entries are required), */
/* fat_creat (open.c) and fat_rename (rename.c).                       */
int allocate_sfn_dir_entry(tFatVolume *V, int Fd, tFatDirEntry *D,
                           char *FileName, DWORD *EntryOffset)
{
  int Res;

  /* Get the absolute directory entry name */
  if (validate_short_file_name(D->Name, FileName))
    return FAT_RET_INVALID_FILENAME;
    
  /* Search for a free directory entry, extending the dir if needed */
  Res = find_empty_dir_entries(V, Fd, 1, EntryOffset);
  if (Res == FAT_RET_DIRECTORY_FULL)
  {
    if (V->Files[Fd].Mode & FAT_FMODE_ROOT) return FAT_RET_DIRECTORY_FULL;
    /* At this point the file is at Eof. We need to extend the directory */
    /* writing a whole cluster filled with zeroes, next we return back   */
    /* to the beginning of that cluster.                                 */
    *EntryOffset = V->Files[Fd].TargetPos;
    /* The following write could fail if disk is full */
    Res = fat_write(V, Fd, NULL,
                    V->Bpb.BPB_BytsPerSec * V->Bpb.BPB_SecPerClus, NULL);
    if (Res) return Res;
  }
  else if (Res) return Res;
  /* Finally write the new directory entry into the free entry */
  V->Files[Fd].TargetPos = *EntryOffset;
  Res = fat_write(V, Fd, (BYTE *) D, 32, NULL);
  if (Res) return Res;
  return 0;
}


/* Allocates 32-bytes directory entries of the specified open directory   */
/* to store a long file name, using D as directory entry for the short    */
/* name alias.                                                            */
/* On success, returns 0 and fills EntryOffset with the byte offset of    */
/* the short name entry, and LfnEntries with the number of entries needed */
/* to store the long name.                                                */
/* Called by fat_creat (open.c) and fat_rename (rename.c).                */
int allocate_lfn_dir_entries(tFatVolume *V, int Fd, tFatDirEntry *D,
                             char *Path, char *FileName, DWORD *EntryOffset,
                             int *LfnEntries, WORD Hint)
{
  char         ShortName[13];
  tFatLfnEntry Slot[20];
  int          Res;
  int          k;
  
  Res = fat_generate_short_name(V, Path, FileName, ShortName, Hint);
  if (Res) return Res;
  
  if (ShortName[0] == 0)
  {
    *LfnEntries = 0;
    return allocate_sfn_dir_entry(V, Fd, D, FileName, EntryOffset);
  }

  /* Generate a 32-byte Directory Entry for the short name alias */
  if (validate_short_file_name(D->Name, ShortName))
    return FAT_RET_INVALID_FILENAME;

  /* Split the long name into 32-byte slots */
  Res = split_lfn(Slot, D, FileName, LfnEntries);
  if (Res) return Res;

  /* Search for NumSlots+1 (LFN entries plus short name entry) */
  /* free directory entries, expanding the dir if needed.      */
  Res = find_empty_dir_entries(V, Fd, *LfnEntries + 1, EntryOffset);
  if (Res == FAT_RET_DIRECTORY_FULL)
  {
    if (V->Files[Fd].Mode & FAT_FMODE_ROOT) return FAT_RET_DIRECTORY_FULL;
    /* At this point the file is at Eof. We need to extend the directory */
    /* writing a whole cluster filled with zeroes, next we return back   */
    /* to the beginning of that cluster.                                 */
    *EntryOffset = V->Files[Fd].TargetPos;
    /* The following write could fail if disk is full */
    Res = fat_write(V, Fd, NULL,
                    V->Bpb.BPB_BytsPerSec * V->Bpb.BPB_SecPerClus, NULL);
    if (Res) return Res;
  }
  else if (Res) return Res;
  /* Finally write the new directory entries into the free entries */
  V->Files[Fd].TargetPos = *EntryOffset;
  for (k = *LfnEntries - 1; k >= 0; k--)
  {
    Res = fat_write(V, Fd, (BYTE *) &Slot[k], 32, NULL);
    if (Res) return Res;
  }
  *EntryOffset = V->Files[Fd].TargetPos;
  Res = fat_write(V, Fd, (BYTE *) D, 32, NULL);
  if (Res) return Res;
  return 0;
}
