/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "init.c" - Mount a FAT volume initializing all data structures *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"
#include "ide.h"

#define MAX_DEVICE_COUNT    4

int           devicecount;
int           defaultdevice = 0; //Drive C: is our default drive
tFatVolume    *devices[MAX_DEVICE_COUNT];

//#define DEBUG

/* The count of sectors occupied by the root directory. Always 0 for FAT32. */
/* Called by first_data_sector and data_clusters.                           */
static /*inline */DWORD root_dir_sectors(tFat32_Bpb *Bpb)
{
  return ((Bpb->BPB_RootEntCnt * 32) + (Bpb->BPB_BytsPerSec - 1)) / Bpb->BPB_BytsPerSec;
}


/* The start of the data region, i.e. the first sector of cluster 2. */
/* Called by fat_mount.                                              */
static /*inline*/ DWORD first_data_sector(tFat32_Bpb *Bpb)
{
  DWORD FATSz;

  if (Bpb->BPB_FATSz16 != 0) FATSz = Bpb->BPB_FATSz16;
                        else FATSz = Bpb->BPB_FATSz32;
  return Bpb->BPB_ResvdSecCnt + (Bpb->BPB_NumFATs * FATSz) + root_dir_sectors(Bpb);
}


/* The count of clusters in the data region of the volume. */
/* Called by determine_fat_type and fat_mount.             */
static /*inline*/ DWORD data_clusters(tFat32_Bpb *Bpb)
{
  DWORD FATSz, TotSec;

  /* For a FAT12/FAT16 volume, BPB_FATSz16 must be non-zero, so we don't */
  /* have to bother to pass a tFat1216_Bpb structure instead.            */
  /* The BPB_FATSz32, only present in a FAT32 BPB, will be used only if  */
  /* if the BPB_FATSz16 is zero, i.e. if the volume is FAT32.            */
  if (Bpb->BPB_FATSz16  != 0) FATSz = Bpb->BPB_FATSz16;
                         else FATSz = Bpb->BPB_FATSz32;
  if (Bpb->BPB_TotSec16 != 0) TotSec = Bpb->BPB_TotSec16;
                         else TotSec = Bpb->BPB_TotSec32;
  return (TotSec - (Bpb->BPB_ResvdSecCnt + (Bpb->BPB_NumFATs * FATSz)
         + root_dir_sectors(Bpb))) / Bpb->BPB_SecPerClus;
}


/* Determines the FAT type according to the count of data clusters */
/* starting at cluster 2, according to Micros*ft specifications.   */
/* The "<" sign and the 4085 and 65525 limits are strongly said by */
/* Micros*ft to be correct.                                        */
/* Returns: FATTYPE_FAT12, FATTYPE_FAT16 or FATTYPE_FAT32.         */
/* Called by fat_mount.                                            */
static int determine_fat_type(tFat32_Bpb *Bpb)
{
  DWORD CountOfClusters = data_clusters(Bpb);

  if (CountOfClusters < 4085) return FATTYPE_FAT12;
  else
  if (CountOfClusters < 65525) return FATTYPE_FAT16;
  else
  return FATTYPE_FAT32;
}


/* The number of free data clusters in the FAT Volume.        */
/* NOTE: We read only the first FAT of the volume (number 0). */
/* On success returns 0 and fills FreeClusters.               */
/* Returns a negative error code on failure.                  */
/* Called by fat_mount.                                       */
static int free_cluster_count(tFatVolume *V, DWORD *FreeClusters)
{
  DWORD Count = 0, k, Value;
  int   Res;
  switch (V->FatType)
  {
    case FATTYPE_FAT12 : for (k = 2; k < V->DataClusters + 2; k++)
                         {
                           Res = fat12_read_entry(V, k, 0, &Value);
                           if (Res) return Res;
                           if (Value == 0) Count++;
                         }
                         *FreeClusters = Count;
                         return 0;
    case FATTYPE_FAT16 : for (k = 2; k < V->DataClusters + 2; k++)
                         {
                           Res = fat16_read_entry(V, k, 0, &Value);
                           if (Res) {
                             return Res;
                           }
                           if (Value == 0) Count++;
                         }
                         *FreeClusters = Count;
                         return 0;
    case FATTYPE_FAT32 : for (k = 2; k < V->DataClusters + 2; k++)
                         {
                           Res = fat32_read_entry(V, k, 0, &Value);
                           if (Res) return Res;
                           if (Value == 0) Count++;
                         }
                         *FreeClusters = Count;
                         return 0;
    default            : /* should not arrive here */
                         return FAT_RET_CRITICAL_ERROR;
  }
}


/* The "free" function for a FAT volume device.                    */
/* Closes all open files, frees the FAT volume and Ops structures. */
/* Returns zero on success, nonzero on failure.                    */
/* This is supposed to be called by the kernel driver engine only. */

static int fat_unmount(void)
{
  int                      k;
  tFatVolume              *V;
  for( int i=0; i < devicecount; i++) {
    V = devices[i];
    if (V != 0) {
      /* Close all files */
      /* FIX ME: Max files and max buffers should be read from CONFIG.SYS! */
      for (k = 0; k < FAT_MAX_FILES; k++) fat_close(V, k);
      #ifdef FATBUFFERS
      for (k = 0; k < FAT_MAX_BUFFERS; k++)
        free(V->Buffer[k].Data);
      #endif
      /* FIX ME: Should free files and buffers arrays */
      free(V);
    }
    devices[i] = 0;
  }
  return 0;
}


/* Partitions signatures that are valid for FAT.  0x00 is allowed when */
/* the partition signature is not available for the hosting device.    */
#define FATPARTITIONS 12
static const BYTE FatPartitions[] = { 0x00, 0x01, 0x04, 0x05, 0x06, 0x0B,
                                      0x0C, 0x0E, 0x0F, 0x1B, 0x1C, 0x1E };

/* Initializes (mounts) a FAT volume */
int fat_mount(Partition* Part)
{
  int                      Res;
  BYTE                    *SecBuf;
  BYTE                     PartSig;
  tFatVolume              *V;

  if (devicecount >= MAX_DEVICE_COUNT) {
    printf("Increase MAX_DEVICE_COUNT to mount this Drive\n");
    return -1;
  }
  PartSig = Part->type;

  printf("Mounting Partition Type %x\n", PartSig);

  /* Check if the specified partition signature is valid for FAT */
  for (Res = 0; Res < FATPARTITIONS; Res++)
    if (PartSig == FatPartitions[Res]) break;
  if (Res == FATPARTITIONS) return -1; /* File system invalid */
  printf("PartSig: %02X, Res: %i\n", PartSig, Res);

//  /* Query the device and read its first sector */
//  Blk = fd32_dev_query(DevName, FD32_DEV_BLOCK);
//  if (Blk == NULL) return -1; /* FIX ME: Error is Invalid device */

//  if (Blk->get_params(Blk->P, &Params)) return -1; /* FIX ME: Error is get_params not supported */

  SecBuf = (char*) malloc(512);  //512 Byte/Sektor IMMER vorrausgesetzt!!!

  unsigned long lba;
  unsigned int c, h, s;
  LBA_to_PCHS(Part->start, &c, &h, &s);
  if (IDE_Read(c, h, s, 1, SecBuf) == 0) {
    free(SecBuf);
    return -1;
  } /* FIX ME: Error is disk read error */

  if (*((WORD*) &SecBuf[510]) != 0xaa55)
  {
    printf("Signature 0xAA55 not found!\n");
    free(SecBuf);
    return -1; /* FIX ME: Error is File system not valid */
  }

  /* Allocate the FAT volume structure and precalculate some data */
  V = (tFatVolume*) malloc(sizeof(tFatVolume));
  memset(V, 0, sizeof(tFatVolume));
  V->PartStart       = Part->start;
  V->DataClusters    = data_clusters((tFat32_Bpb *) SecBuf);
  V->FirstDataSector = first_data_sector((tFat32_Bpb *) SecBuf);
  V->FatType         = determine_fat_type((tFat32_Bpb *) SecBuf);

  printf("Fat Type: %i\n", V->FatType);
  printf("First Data Sector: %lu\n", V->FirstDataSector);
  printf("Data Clusters: %lu\n", V->DataClusters);

  /* Initialize the volume's BPB, using also the FSInfo if FAT32. */
  if (V->FatType == FATTYPE_FAT32)
  {
    tFat32_FSInfo *FSInfo;

    memcpy(&V->Bpb, SecBuf, sizeof(tFat32_Bpb));
    if (V->Bpb.BPB_FSVer > 0x0000)
    {
      printf("Unknown FAT32 version!\n");
      free(SecBuf);
      free(V);
      return -1; /* FIX ME: Error is File system not valid */
    }
    /* Read the FAT32 FSInfo Sector structure */
    FSInfo = (tFat32_FSInfo*) malloc(512);
    LBA_to_PCHS(V->Bpb.BPB_FSInfo+Part->start, &c, &h, &s);
    if (IDE_Read(c, h, s, 1, (char*) FSInfo) == 0) {
      free(V);
      free(SecBuf);
      free(FSInfo);
      return -1; /* FIX ME: Error is disk read error */
    }
    if (FSInfo->LeadSig != 0x41615252)
    {
      printf("Invalid FAT32 FSInfo FSI_LeadSig! %lu\n", FSInfo->LeadSig);
      free(V);
      free(SecBuf);
      free(FSInfo);
      return -1; /* FIX ME: Error is File system not valid */
    }
    if (FSInfo->StrucSig != 0x61417272)
    {
      printf("Invalid FAT32 FSInfo FSI_StrucSig!\n");
      free(V);
      free(SecBuf);
      free(FSInfo);
      return -1; /* FIX ME: Error is File system not valid */
    }
    V->CurDirCluster  = V->Bpb.BPB_RootClus;
    V->FSI_Free_Count = FSInfo->Free_Count;
    V->FSI_Nxt_Free   = FSInfo->Nxt_Free;
    free(FSInfo);
  }
  else /* if not FAT32 */
  {
    /* Clear the unused tFat32_Bpb fields                              */
    memset(&V->Bpb, 0, sizeof(tFat32_Bpb));
    /* The first 36 bytes of tFat1216_Bpb and tFat32_Bpb are the same  */
    memcpy(&V->Bpb, SecBuf, 36);
    /* The 26 bytes of tFat1216_Bpb starting at offset 36 are the same */
    /* as the 26 bytes of tFat32_Bpb starting at offset 64             */
    memcpy(&V->Bpb.BS_DrvNum, SecBuf + 36, 26);
    V->CurDirCluster  = 0;
    V->FSI_Free_Count = 0xFFFFFFFF;
    V->FSI_Nxt_Free   = 0xFFFFFFFF;
  }
  // [KG] Adjust Buffer, to contain no more than one cluster
  V->SectorsInBuffer = V->Bpb.BPB_SecPerClus;
  if (V->SectorsInBuffer > 8) {
    V->SectorsInBuffer = 8;
  }
  free(SecBuf);

  #ifdef FATBUFFERS
  /* Initialize volume's buffers */
  {
    int k;
    /* FIX ME: The buffers number should be read from CONFIG.SYS! */
    for (k = 0; k < FAT_MAX_BUFFERS; k++)
    {
      V->Buffer[k].StartingSector = 0xFFFFFFFF;
      V->Buffer[k].LastAccess     = 0;
      V->Buffer[k].Flags          = 0;
      V->Buffer[k].Data           = (BYTE *) malloc(512*V->SectorsInBuffer);
    }
    V->BufferAccess = 0;
  }
  #endif

  #ifdef DEBUG
  printf("Scanning FAT for free clusters...\n");
  #endif
  /* If the count of free clusters is not available or it's greater    */
  /* than the count of clusters of the volume (DataClusters + 1) we    */
  /* need to compute it by scanning the FAT.                           */
  if ((V->FSI_Free_Count == 0xFFFFFFFF) ||
      (V->FSI_Free_Count > V->DataClusters + 1))
  {
    printf("Detecting free diskspace, please wait...\n");  
    Res = free_cluster_count(V, &V->FSI_Free_Count);
    if (Res) return Res;
  }
  printf("%lu clusters available\n", V->FSI_Free_Count);
  /* Set the root directory as current directory */
  V->CurDirPath[0] = '\\';
  #ifdef DEBUG
  printf("FAT Volume Initialized.\n");
  #endif

  V->DevName[0] = 'C'+devicecount;
  V->DevName[1] = '\0';
//  strcpy(V->DevName, "FATxx Partition");
//  fd32_list_devices();
  devices[devicecount] = V;
  devicecount++;
  return 0;
}

